// Author: Pete Kirkham
//
// quad store.
//
// simple file-reader

#ifndef _LEAP_FROG_FILE_READER_H_
#define _LEAP_FROG_FILE_READER_H_

#include    "utils/threading.h"
#include    "io/reader_base.h"

#define _XOPEN_SOURCE 600

#include <fcntl.h>

#include <pthread.h>

class leap_frog_file_reader
{
public:
    inline leap_frog_file_reader (int file) : _file(file), _block_start(-BLOCK_SIZE), _fail(false), _index(BLOCK_SIZE)
    {
        _out_buf[0] = 0;
        
        typedef void * (*pthread_start_routine)(void *);

        pthread_create(&_leap_thread, 0, reinterpret_cast<pthread_start_routine>(&leap_frog_file_reader::do_run), reinterpret_cast<void*>(this));
    }
    
    static void do_run (void* that)
    {
        reinterpret_cast<leap_frog_file_reader*>(that)->read_blocks_as_required();
    }
    
    void read_blocks_as_required ()
    {
        char    alloc_buf0[BLOCK_SIZE];
        char    alloc_buf1[BLOCK_SIZE];
        char*   internal_buffer(alloc_buf0);
        
        for (;;)
        {
            // read from file
            int     nread(::read(_file, internal_buffer, BLOCK_SIZE));
            
            // wait for a request
            _data_required.wait_for();

            // switch buffers
            _read_buf = internal_buffer;
            
            internal_buffer = internal_buffer == alloc_buf1 ? alloc_buf0 : alloc_buf1;
            
            // set flags
            if (nread != BLOCK_SIZE)
            {
                set_failed();
                break;
            }

            // signal data available
            _data_available.signal();
        }
    }

    inline size_t  total_read ()
    {
        return _block_start + _index;
    }

    inline bool    failed ()
    {
        return _fail;
    }

    inline bool    ok ()
    {
        return !_fail;
    }

    inline bool    seek (int offset)
    {
        if (lseek(_file, offset, SEEK_SET) != (off_t)-1)
        {
            _block_start = offset;
            _index = BLOCK_SIZE;
            _data_required.signal();

            return true;
        }
        else
        {
            return false;
        }
    }

    inline const char*   read    (size_t length)
    {
        if (length >= BLOCK_SIZE)
        {
            length = BLOCK_SIZE - 1;
        }

        if (_index + length >= BLOCK_SIZE)
        {
            size_t i2(0);

            for (size_t i0(_index); i0 < BLOCK_SIZE; ++i0, ++i2)
            {
                _out_buf[i2] = _read_buf[i0];
            }

            // wait for leap_thread
            _data_required.signal();
            _data_available.wait_for();
            
            if (!_fail)
            {
                for (size_t i0(0); i2 < length; ++i0, ++i2)
                {
                    _out_buf[i2] = _read_buf[i0];
                }

                _index += length - BLOCK_SIZE;

                _block_start += BLOCK_SIZE;
            }
        }
        else
        {
            for (size_t i0(_index), i2(0); i2 < length; ++i0, ++i2)
            {
                _out_buf[i2] = _read_buf[i0];
            }

            _index += length;
        }

        _out_buf[length] = 0;

        return _out_buf;
    }

    inline int   read    ()
    {
        if (_index < BLOCK_SIZE)
        {
            int result(_read_buf[_index]);
            ++_index;
            return result;
        }
        else
        {
            const char* buf(read(1));

            return _fail ? -1  : buf[0];
        }
    }

    inline char*   last ()
    {
        return _out_buf;
    }

protected:
    void set_failed ()
    {
        _fail = true;
        _data_available.signal();
    }

private:
    int         _file;
    bool        _fail;
    char*       _read_buf;  // raw read buffer
    char        _out_buf[BLOCK_SIZE];  // result of read(n)
    size_t      _block_start;
    size_t      _index;

    pthread_t   _leap_thread;

    condition   _data_required;
    condition   _data_available;
};

#endif // _LEAP_FROG_FILE_READER_H_
